/*
 * Copyright (c) 2013, 2014 - University of Virginia 
 *
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 */

#include "coverage.h"
#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <limits.h>
#include <string>
#include <fstream>
#include <libgen.h>


using namespace std;
using namespace libIRDB;

void trim(string& str)
{
  string::size_type pos = str.find_last_not_of(' ');
  if(pos != string::npos) {
    str.erase(pos + 1);
    pos = str.find_first_not_of(' ');
    if(pos != string::npos) str.erase(0, pos);
  }
  else str.erase(str.begin(), str.end());
}

enum STR2NUM_ERROR { STR2_SUCCESS, STR2_OVERFLOW, STR2_UNDERFLOW, STR2_INCONVERTIBLE };

//TODO: what if the string represents a negative number? Currently
//the number will be translated into an unsigned int. I could make this
//and incovertible situation. 
STR2NUM_ERROR str2uint (unsigned int &i, char const *s, int base=0)
{
	char *end;
	unsigned long  l;
	errno = 0;
	l = strtoul(s, &end, base);
	if ((errno == ERANGE && l == ULONG_MAX) || l > UINT_MAX) {
		return STR2_OVERFLOW;
	}
	if (*s == '\0' || *end != '\0') {
		return STR2_INCONVERTIBLE;
	}
	i = l;

	return STR2_SUCCESS;
}


void coverage::parse_coverage_file(ifstream &coverage_file)
{
	string line;
	while(coverage_file.is_open() && std::getline(coverage_file,line))
	{
		trim(line);

		if(line.empty())
				continue;

		string file,addr;

		istringstream iss(line);

		getline(iss,file,'+');
		assert(!file.empty());
		getline(iss,addr,'+');

		//if addr is empty, assume the entry for file
		//is the addr
		trim(addr);
		if(addr.empty())
		{
			addr = file;
			file = "a.ncexe";
		}


		if(coverage_map.find(file) == coverage_map.end())
		{
			file_coverage fc;
			fc.file = file;
			coverage_map[file]=fc;
		}			

		
		unsigned int uint_addr;
		assert(str2uint(uint_addr,addr.c_str())==STR2_SUCCESS);

		coverage_map[file].coverage[uint_addr]=uint_addr;
	}
}

file_coverage* coverage::find_file_coverage(string url)
{
	for(map<string, file_coverage >::iterator it=coverage_map.begin();
		it!=coverage_map.end(); ++it)
	{
		string key = it->first;

		//TODO: for now taking the base name, since that is how
		//the shared objects are registered, but this might cause
		//issues if shared objects have the same name. 

		key=string(basename((char*)key.c_str()));

		if(key.empty())
			continue;

		if(url.find(key)!=string::npos)
		{
			return &(it->second);
		}
	}

	return NULL;
}

void coverage::print_coverage_for_file(file_coverage *fc, FileIR_t *fileirp, ofstream &out_file)
{
	for(
		set<Function_t*>::const_iterator func_it=fileirp->GetFunctions().begin();
		func_it!=fileirp->GetFunctions().end();
		++func_it
		)
	{
		Function_t *func = *func_it;
		if(func==NULL)
			continue;

		unsigned int total_ins = func->GetInstructions().size();
		unsigned int covered_ins = 0;

		for(
			set<Instruction_t*>::const_iterator it=func->GetInstructions().begin();
			it!=func->GetInstructions().end();
			++it
			)
		{
			Instruction_t* instr = *it;

			if(instr==NULL || instr->GetAddress()==NULL)
				continue;

			unsigned int addr = (unsigned int) instr->GetAddress()->GetVirtualOffset();
			if(fc->coverage.find(addr)!=fc->coverage.end())
				covered_ins++;
		}

		double cov_percent = ((double)covered_ins)/((double)total_ins);

		out_file<<fc->file<<"+"<<func->GetName()<<" "<<cov_percent<<" "<<covered_ins<<" "<<total_ins<<endl;

	}
}

void coverage::print_function_coverage_file(libIRDB::VariantID_t *vidp,std::ofstream &out_file)
{
	assert(vidp);

	for(set<File_t*>::iterator it=vidp->GetFiles().begin();
		it!=vidp->GetFiles().end();
		++it
		)
	{
		File_t* this_file=*it;
		assert(this_file);

		// read the db  
		FileIR_t *fileirp=new FileIR_t(*vidp,this_file);
		assert(fileirp);

		//NULL fc should only happen if the coverage file is empty. 
		//TODO: add extra sanity checks to make sure that null only happens under these conditions. 
		file_coverage* fc = find_file_coverage(fileirp->GetFile()->GetURL());
		if(fc!=NULL)
		{
			print_coverage_for_file(fc, fileirp, out_file);
		}

		delete fileirp;
	}
}
